#include <stdio.h>
#include <stdlib.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include <brick.h>
#include "lua-mod.h"
#include "lua-mod-frame.h"






/* retrieve frame info .. */
int wrap_frame_info(lua_State *l)
{
	/* a lookup table, to give us the actual named mode .. */
	const char *modes[] = { "none", "rgba", "rgb", "hl", "sl", "br", "ct", "sat", "displ", "convo", "lut", "xor" };
	frame *f;

	/* .. and the results! */
	int w, h, mode;


	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	/* get the frame info */
	if( frame_info(f, &w, &h, &mode) == ERR )
		return luaL_error(l, "Invalid frame ");
	else
		{
			/* and prepare the results */
			lua_pushinteger(l, w);
			lua_pushinteger(l, h);
			lua_pushstring(l, modes[mode]);
			return 3;
		}

}





/* create a frame */
int wrap_frame_create(lua_State *l)
{
	int i;

	/* data to retrieve from the lua arguments */
	const char *mode;
	int w,h;
	const unsigned char *data; /* stored in a lua string */
	size_t len;                   /* length of data */

	/* also cumbersome - variables to read and parse convolution kernel or lookup table */
	convolution ck;
	lut ldata;

	/* frame pointer */
	frame *f;


	/* check for some args */
	mode = lua_tostring(l, 1);
	w = lua_tonumber(l, 2);
	h = lua_tonumber(l, 3);

	/* and parse out based on mode */
	if( !strcmp(mode, "none") )
		f = frame_create(FRAME_NONE, w, h, NULL, NULL);
	else
		{
			/* one of the non-null sprite types */
			data = lua_tolstring(l, 4, &len);

			if( !strcmp(mode, "rgba") )
				f = frame_create(FRAME_RGBA, w, h, data, NULL);
			else if( !strcmp(mode, "rgb") )
				f = frame_create(FRAME_RGB, w, h, data, NULL);
			else if( !strcmp(mode, "br") )
				f = frame_create(FRAME_BR, w, h, data, NULL);
			else if( !strcmp(mode, "ct") )
				f = frame_create(FRAME_CT, w, h, data, NULL);
			else if( !strcmp(mode, "hl") )
				f = frame_create(FRAME_HL, w, h, data, NULL);
			else if( !strcmp(mode, "sl") )
				f = frame_create(FRAME_SL, w, h, data, NULL);
			else if( !strcmp(mode, "sat") )
				f = frame_create(FRAME_SAT, w, h, data, NULL);
			else if( !strcmp(mode, "displ") )
				f = frame_create(FRAME_DISPL, w, h, (short *)data, NULL);
			else if( !strcmp(mode, "convo") )
			  {
					/*
					 * retrieve the convolution-kernel data from the arguments thus:
					 * w h [ k1 k2 k3 .. kn ] divisor offset
					 */

					/* first fetch the kernel width and height */
					ck.kw = lua_tonumber(l, 5);
					ck.kh = lua_tonumber(l, 6);

					if( ck.kw > MAX_CK_SIZE || ck.kh > MAX_CK_SIZE )
						return luaL_error(l,  "Convolution kernel too large ");

					/* convolution kernel table is in the stack at index 7 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 7); i++ )
						{
						  ck.kernel[i] = (char)lua_tonumber(l, -1);
							lua_pop(l, 1);
						}

					ck.divisor = lua_tonumber(l, 8);
					ck.offset = lua_tonumber(l, 9);

					if( i != ck.kw * ck.kh )
					  return luaL_error(l, "Incorrect amount of kernel data provided ");

					/* at last, add the convolution kernel */
					f = frame_create(FRAME_CONVO, w, h, data, &ck);

			  }
			else if( !strcmp(mode, "lut") )
			  {
					/* retrieve the lookup table data from three provided tables */

					/* red table is in the stack at index 5 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 5) && i < RGB_RANGE; i++ )
						{
							ldata.r[i] = (unsigned char)lua_tonumber(l, -1);
							lua_pop(l, 1);
					  }

					/* green table is in the stack at index 6 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 6) && i < RGB_RANGE; i++ )
						{
							ldata.g[i] = (unsigned char)lua_tonumber(l, -1);
							lua_pop(l, 1);
					  }

					/* blue table is in the stack at index 7 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 7) && i < RGB_RANGE; i++ )
						{
							ldata.b[i] = (unsigned char)lua_tonumber(l, -1);
							lua_pop(l, 1);
					  }

					/* at last, add the convolution kernel */
					f = frame_create(FRAME_LUT, w, h, data, &ldata);
			  }
			else if( !strcmp(mode, "xor") )
				f = frame_create(FRAME_XOR, w, h, data, NULL);
			else
			  return luaL_error(l, "Unknown frame type ");
		}


	/* and finished */
	if( !f )
		return luaL_error(l, "Invalid frame ");
	else
		{
			lua_pushlightuserdata(l, f);
			return 1;
		}

}






/* deleting a frame .. */
int wrap_frame_delete(lua_State *l)
{
	frame *f;

	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	frame_delete(f);
	return 0;

}




/* copying a frame .. */
int wrap_frame_copy(lua_State *l)
{
	frame *f;
	int refcopy;

	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	lua_pushlightuserdata(l, frame_copy(f));
	return 1;

}




/* set frame offset .. */
int wrap_frame_set_offset(lua_State *l)
{
	frame *f;
	int x, y;

	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	x = lua_tonumber(l, 2);
	y = lua_tonumber(l, 3);

	frame_set_offset(f, x, y);
	return 0;

}



/* set the pixel mask */
int wrap_frame_set_mask(lua_State *l)
{
	/* the frame in question .. */
	frame *f;

	/* pixel data buffer */
	size_t len;
	const unsigned char *data;

	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	/* check that there is the right amount of data .. */
	data = lua_tolstring(l, 2, &len);
	if( len != f->w * f->h )
		return luaL_error(l, "Amount of pixel mask data doesn't match frame size ");

	frame_set_mask(f, data);
	return 0;
}





/* set the pixel mask from a source frame */
int wrap_frame_set_mask_from(lua_State *l)
{
	frame *f, *src;

	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	src = lua_touserdata(l, 2);
	if( !src )
		return luaL_error(l, "Invalid frame ");

	frame_set_mask_from(f, src);
	return 0;
}




/* slice up a frame */
int wrap_frame_slice(lua_State *l)
{
	frame *f, *res;
	int x, y, w, h;

	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	x = lua_tonumber(l, 2);
	y = lua_tonumber(l, 3);
	w = lua_tonumber(l, 4);
	h = lua_tonumber(l, 5);

	res = frame_slice(f, x, y, w, h);
	if( !res )
		return luaL_error(l, "Invalid frame ");
	else
		{
			lua_pushlightuserdata(l, res);
			return 1;
		}

}




/* convert a frame */
int wrap_frame_convert(lua_State *l)
{
	/* the source frame and requested mode */
	frame *f, *res;
	const char *mode;

	/* also cumbersome - variables to read and parse convolution kernel or lookup table */
	convolution ck;
	lut ldata;

	int i;


	/* retrieve the frame and the desired conversion type */
	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	mode = lua_tostring(l, 2);


	/* which frame type? */
	if( !strcmp(mode, "rgb") )
		res = frame_convert(f, FRAME_RGB, NULL);
	else if( !strcmp(mode, "rgba") )
		res = frame_convert(f, FRAME_RGBA, NULL);
	else if( !strcmp(mode, "hl") )
		res = frame_convert(f, FRAME_HL, NULL);
	else if( !strcmp(mode, "sl") )
		res = frame_convert(f, FRAME_SL, NULL);
	else if( !strcmp(mode, "br") )
		res = frame_convert(f, FRAME_BR, NULL);
	else if( !strcmp(mode, "ct") )
		res = frame_convert(f, FRAME_CT, NULL);
	else if( !strcmp(mode, "sat") )
		res = frame_convert(f, FRAME_SAT, NULL);
	else if( !strcmp(mode, "convo") )
		{
			/*
			 * retrieve the convolution-kernel data from the arguments thus:
			 * w h [ k1 k2 k3 .. kn ] divisor offset
			 */

			/* first fetch the kernel width and height */
			ck.kw = lua_tonumber(l, 3);
			ck.kh = lua_tonumber(l, 4);

			if( ck.kw > MAX_CK_SIZE || ck.kh > MAX_CK_SIZE )
				return luaL_error(l,  "Convolution kernel too large ");

			/* convolution kernel table is in the stack at index 5 */
			lua_pushnil(l);
			for( i=0; lua_next(l, 5); i++ )
				{
					ck.kernel[i] = (char)lua_tonumber(l, -1);
					lua_pop(l, 1);
				}

			ck.divisor = lua_tonumber(l, 6);
			ck.offset = lua_tonumber(l, 7);

			if( i != ck.kw * ck.kh )
				return luaL_error(l, "Incorrect amount of kernel data provided ");

			/* at last, add the convolution kernel */
			res = frame_convert(f, FRAME_CONVO, &ck);

		}
	else if( !strcmp(mode, "lut") )
		{
			/* retrieve the lookup table data from three provided tables */

			/* red table is in the stack at index 3 */
			lua_pushnil(l);
			for( i=0; lua_next(l, 3) && i < RGB_RANGE; i++ )
				{
					ldata.r[i] = (unsigned char)lua_tonumber(l, -1);
					lua_pop(l, 1);
				}

			/* green table is in the stack at index 4 */
			lua_pushnil(l);
			for( i=0; lua_next(l, 4) && i < RGB_RANGE; i++ )
				{
					ldata.g[i] = (unsigned char)lua_tonumber(l, -1);
					lua_pop(l, 1);
				}

			/* blue table is in the stack at index 5 */
			lua_pushnil(l);
			for( i=0; lua_next(l, 5) && i < RGB_RANGE; i++ )
				{
					ldata.b[i] = (unsigned char)lua_tonumber(l, -1);
					lua_pop(l, 1);
				}

			/* at last, add the convolution kernel */
			res = frame_convert(f, FRAME_LUT, &l);
		}
	else if( !strcmp(mode, "xor") )
		res = frame_convert(f, FRAME_XOR, NULL);
	else
		return luaL_error(l, "Unknown frame type ");

	/* error checking .. */
	if( !res )
		return luaL_error(l, "Invalid frame or requested new type ");
	else
		{
			lua_pushlightuserdata(l, res);
			return 1;
		}

}




/* create an effect frame .. */
int wrap_frame_effect(lua_State *l)
{
	frame *f;
	const char *mode;

	/* options for drop shadow */
	int x, y, blur;
	int r, g, b;
	color c;


	f = lua_touserdata(l, 1);
	if( !f )
		return luaL_error(l, "Invalid frame ");

	mode = lua_tostring(l, 2);

	/* which effect type? */
	if( !strcmp(mode, "dropshadow") )
		{
			/* read the position and blur options */
			x = lua_tonumber(l, 3);
			y = lua_tonumber(l, 4);
			blur = lua_tonumber(l, 5);

			/* and the drop shadow from the table at index 6 */
			if( !lua_istable(l, 6) )
				luaL_error(l, "Drop shadow color isn't an rgb triplet ");

			lua_pushnil(l);
			lua_next(l, 6);
			r = lua_tonumber(l, -1); c.r = (unsigned char)r;
			lua_pop(l, 1);

			lua_next(l, 6);
			g = lua_tonumber(l, -1); c.g = (unsigned char)g;
			lua_pop(l, 1);

			lua_next(l, 6);
			b = lua_tonumber(l, -1); c.b = (unsigned char)b;
			lua_pop(l, 1);

			/* and create the drop shadow */
			lua_pushlightuserdata(l, frame_effect(f, FRAME_EFFECT_DROP_SHADOW, x, y, blur, &c));
			return 1;
		}
	else
		return luaL_error(l, "Unknown frame effect ");

}




/* we have been requested to load a frame from a file */
int wrap_frame_from_disk(lua_State *l)
{
	/* source filename */
	const char *file;

	/* result */
	frame *res;

	file = lua_tostring(l, 1);
	res = frame_from_disk(file);

	/* check return code */
	if( !res )
		return luaL_error(l, "Bad file ");
	else
		{
			lua_pushlightuserdata(l, res);
			return 1;
		}

}




/* we have been requested to load a frame from a memory buffer */
int wrap_frame_from_buffer(lua_State *l)
{
	/* source buffer */
	size_t len;
	const unsigned char *data;

	/* result */
	frame *res;

	data = lua_tolstring(l, 1, &len);
	res = frame_from_buffer(len, data);

	/* check return code */
	if( !res )
		return luaL_error(l, "Bad file ");
	else
		{
			lua_pushlightuserdata(l, res);
			return 1;
		}

}
