#include <stdio.h>
#include <stdlib.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include <brick.h>
#include "lua-mod.h"
#include "lua-mod-inspect.h"







/* inspect the map tiles surrounding the given sprite */
int wrap_inspect_adjacent_tiles(lua_State *l)
{
	/* the sprite and map */
	sprite *sprite;
	const char *dir;
	map *map;

	/* results */
	map_fragment res;
	int i;


	/* get the sprite */
	sprite = lua_touserdata(l, 1);
	if( !sprite )
		return luaL_error(l, "Invalid sprite ");

	/* the direction */
	dir = lua_tostring(l, 2);

	/* and the map */
	map = lua_touserdata(l, 3);
	if( !map )
		return luaL_error(l, "Invalid map ");


	/* call the c function */
	if( !strcmp(dir, "nw") )
		inspect_adjacent_tiles(map, sprite, INSPECT_NW, &res);
	else if( !strcmp(dir, "n") )
		inspect_adjacent_tiles(map, sprite, INSPECT_N, &res);
	else if( !strcmp(dir, "ne") )
		inspect_adjacent_tiles(map, sprite, INSPECT_NE, &res);
	else if( !strcmp(dir, "e") )
		inspect_adjacent_tiles(map, sprite, INSPECT_E, &res);
	else if( !strcmp(dir, "se") )
		inspect_adjacent_tiles(map, sprite, INSPECT_SE, &res);
	else if( !strcmp(dir, "s") )
		inspect_adjacent_tiles(map, sprite, INSPECT_S, &res);
	else if( !strcmp(dir, "sw") )
		inspect_adjacent_tiles(map, sprite, INSPECT_SW, &res);
	else if( !strcmp(dir, "w") )
		inspect_adjacent_tiles(map, sprite, INSPECT_W, &res);

	/* and build the results table */
	lua_newtable(l);
	for( i=0; i < res.w * res.h; i++ )
		{
			lua_pushinteger(l, i+1);
			lua_pushinteger(l, *(res.tiles + i));
			lua_settable(l, -3);
		}

	/* important!  only free it if we actually got a result .. */
	if( res.tiles )
		free(res.tiles);

	return 1;

}




/* inspect the map tiles obscured by the given sprite */
int wrap_inspect_obscured_tiles(lua_State *l)
{
	/* the sprite and map */
	sprite *sprite;
	map *map;

	/* results */
	map_fragment res;
	int i, j, ofs;


	/* get the sprite */
	sprite = lua_touserdata(l, 1);
	if( !sprite )
		return luaL_error(l, "Invalid sprite ");

	/* and the map */
	map = lua_touserdata(l, 2);
	if( !map )
		return luaL_error(l, "Invalid map ");


	/* call the c function */
	inspect_obscured_tiles(map, sprite, &res);

	lua_pushnumber(l, res.w);
	lua_pushnumber(l, res.h);

	/* prepare the results table */
	lua_newtable(l);
	for( i=0; i < res.w * res.h; i++ )
		{
			lua_pushinteger(l, i+1);
			lua_pushinteger(l, *(res.tiles + i));
			lua_settable(l, -3);
		}


	/* important!  only free it if we actually got a result .. */
	if( res.tiles )
		free(res.tiles);

	return 3;

}




/* check line-of-sight visibility between two sprites on the given map */
int wrap_inspect_line_of_sight(lua_State *l)
{
	sprite *sprite, *target;
	map *map;
	int xofs, yofs, dist;

	sprite = lua_touserdata(l, 1);
	if( !sprite )
		return luaL_error(l, "Invalid sprite ");

	xofs = lua_tonumber(l, 2);
	yofs = lua_tonumber(l, 3);
	dist = lua_tonumber(l, 4);

	target = lua_touserdata(l, 5);
	if( !target )
		return luaL_error(l, "Invalid target ");

	map = lua_touserdata(l, 6);
	if( !map )
		return luaL_error(l, "Invalid map ");

	/* and run the test, and we're done! */
	lua_pushboolean(l, inspect_line_of_sight(map, sprite, xofs, yofs, dist, target));

}






/* return a list of sprites, drawn from the given list, within the given frame */
int wrap_inspect_in_frame(lua_State *l)
{
	/* the sprite list pointer and frame */
	list *my_l;
	box range;

	/* results */
	list *res;
	iterator iter;


	my_l = lua_touserdata(l, 1);
	if( !my_l )
		return luaL_error(l, "Invalid list ");

	range.x1 = lua_tonumber(l, 2);
	range.y1 = lua_tonumber(l, 3);
	range.x2 = lua_tonumber(l, 4);
	range.y2 = lua_tonumber(l, 5);
	res = inspect_in_frame(my_l, &range);

	/* Prepare the table with the results */
	lua_newtable(l);

	iterator_start(iter, res);
	while( iterator_data(iter) )
		{
			lua_pushinteger(l, iterator_ct(iter)+1);
			lua_pushlightuserdata(l, iterator_data(iter));
			lua_settable(l, -3);

			iterator_next(iter);
		}

	/* free the result list */
	list_delete(res);

	/* and return the table */
	return 1;

}





/* return a list of sprites, drawn from the given list, within the specified distance of the given point */
int wrap_inspect_near_point(lua_State *l)
{
	/* the sprite list, location, and distance */
	list *my_l;
	int x, y, dist;

	/* results */
	list *res;
	iterator iter;


	my_l = lua_touserdata(l, 1);
	if( !my_l )
		return luaL_error(l, "Invalid list ");

	x = lua_tonumber(l, 2);
	y = lua_tonumber(l, 3);
	dist = lua_tonumber(l, 4);
	res = inspect_near_point(my_l, x, y, dist);

	/* Prepare the table with the results */
	lua_newtable(l);

	iterator_start(iter, res);
	while( iterator_data(iter) )
		{
			lua_pushinteger(l, iterator_ct(iter)+1);
			lua_pushlightuserdata(l, iterator_data(iter));
			lua_settable(l, -3);

			iterator_next(iter);
		}

	/* free the result list */
	list_delete(res);

	/* and return the table */
	return 1;

}

