#include <stdio.h>
#include <stdlib.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include <brick.h>
#include "lua-mod.h"
#include "lua-mod-tile.h"




/* creating a new tile .. */
int wrap_tile_create(lua_State *l)
{
	lua_pushlightuserdata(l, tile_create());
	return 1;
}




/* deleting a tile .. */
int wrap_tile_delete(lua_State *l)
{
	tile *tile;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	tile_delete(tile);
	return 0;
}




/* set the animation type for the given tile */
int wrap_tile_anim_type(lua_State *l)
{
	/* possible types */
	const char *type, *types[] = { "still", "fwd", "rev", "pp" };
	int idx;
	tile *tile;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	if( lua_gettop(l) == 2 )
		{
			/* set the animation type */
			type = lua_tostring(l, 2);

			if( !strcmp(type, "off")  )
				tile_set_anim_type(tile, ANIMATE_OFF);
			else if( !strcmp(type, "fwd") )
				tile_set_anim_type(tile, ANIMATE_FWD);
			else if( !strcmp(type, "rev") )
				tile_set_anim_type(tile, ANIMATE_REV);
			else if( !strcmp(type, "pp") )
				tile_set_anim_type(tile, ANIMATE_PP);
			else
				return luaL_error(l, "Unknown animation type ");

			return 0;
		}
	else
		{
			if( tile_get_anim_type(tile, &idx) == ERR )
				return luaL_error(l, "Invalid tile id ");
			else
				{
					lua_pushstring(l, types[idx]);
					return 1;
				}

		}

}




/* set the collision mode for the given tile */
int wrap_tile_collides(lua_State *l)
{
	/* collision modes */
	const char *mode, *modes[] = { "off", "box", "pixel" };
	int idx;
	tile *tile;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	if( lua_gettop(l) == 2 )
		{
			/* set the collision type */
			mode = lua_tostring(l, 2);

			if( !strcmp(mode, "off") )
				tile_set_collides(tile, COLLISION_OFF);
			else if( !strcmp(mode, "box") )
				tile_set_collides(tile, COLLISION_BOX);
			else if( !strcmp(mode, "pixel") )
				tile_set_collides(tile, COLLISION_PIXEL);
			else
				return luaL_error(l, "Unknown collision mode ");

			return 0;
		}
	else
		{
			/* read and return the name of the animation type */
			if( tile_get_collides(tile, &idx) == ERR )
				return luaL_error(l, "Invalid tile id ");
			else
				{
					lua_pushstring(l, modes[idx]);
					return 1;
				}

		}

}




/* set a tile frame .. */
int wrap_tile_add_frame(lua_State *l)
{
	tile *tile;
	frame *frame;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	frame = lua_touserdata(l, 2);
	if( !frame )
		return luaL_error(l, "Invalid frame ");

	lua_pushinteger(l, tile_add_frame(tile, frame));
	return 1;
}




/* add a frame to a tile */
int wrap_tile_add_frame_data(lua_State *l)
{
	/* tile pointer */
	tile *tile;

	/* data to retrieve from the lua arguments */
	const char *mode;
	int w,h;
	const unsigned char *data; /* stored in a lua string */
	size_t len;                /* length of data */

	/* also cumbersome - variables to read and parse convolution kernel or lookup table */
	convolution ck;
	lut ldata;

	int i;


	/* the result */
	int res;

	/* check for some args */
	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	mode = lua_tostring(l, 2);
	w = lua_tonumber(l, 3);
	h = lua_tonumber(l, 4);

	/* and parse out based on mode */
	if( !strcmp(mode, "none") )
		res = tile_add_frame_data(tile, FRAME_NONE, w, h, NULL, NULL);
	else
		{
			/* one of the non-null sprite types */
			data = lua_tolstring(l, 5, &len);

			if( !strcmp(mode, "rgba") )
				res = tile_add_frame_data(tile, FRAME_RGBA, w, h, data, NULL);
			else if( !strcmp(mode, "rgb") )
				res = tile_add_frame_data(tile, FRAME_RGB, w, h, data, NULL);
			else if( !strcmp(mode, "br") )
				res = tile_add_frame_data(tile, FRAME_BR, w, h, data, NULL);
			else if( !strcmp(mode, "ct") )
				res = tile_add_frame_data(tile, FRAME_CT, w, h, data, NULL);
			else if( !strcmp(mode, "hl") )
				res = tile_add_frame_data(tile, FRAME_HL, w, h, data, NULL);
			else if( !strcmp(mode, "sl") )
				res = tile_add_frame_data(tile, FRAME_SL, w, h, data, NULL);
			else if( !strcmp(mode, "sat") )
				res = tile_add_frame_data(tile, FRAME_SAT, w, h, data, NULL);
			else if( !strcmp(mode, "displ") )
				res = tile_add_frame_data(tile, FRAME_DISPL, w, h, data, NULL);
			else if( !strcmp(mode, "convo") )
			  {
					/*
					 * retrieve the convolution-kernel data from the arguments thus:
					 * w h [ k1 k2 k3 .. kn ] divisor offset
					 */

					/* first fetch the kernel width and height */
					ck.kw = lua_tonumber(l, 6);
					ck.kh = lua_tonumber(l, 7);

					if( ck.kw > MAX_CK_SIZE || ck.kh > MAX_CK_SIZE )
						return luaL_error(l,  "Convolution kernel too large ");

					/* convolution kernel table is in the stack at index 7 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 8); i++ )
						{
						  ck.kernel[i] = (char)lua_tonumber(l, -1);
							lua_pop(l, 1);
						}

					ck.divisor = lua_tonumber(l, 9);
					ck.offset = lua_tonumber(l, 10);

					if( i != ck.kw * ck.kh )
					  return luaL_error(l, "Incorrect amount of kernel data provided ");

					/* at last, add the convolution kernel */
					res = tile_add_frame_data(tile, FRAME_CONVO, w, h, data, &ck);

			  }
			else if( !strcmp(mode, "lut") )
			  {
					/* retrieve the lookup table data from three provided tables */

					/* red table is in the stack at index 5 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 6) && i < RGB_RANGE; i++ )
						{
							ldata.r[i] = (unsigned char)lua_tonumber(l, -1);
							lua_pop(l, 1);
					  }

					/* green table is in the stack at index 6 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 7) && i < RGB_RANGE; i++ )
						{
							ldata.g[i] = (unsigned char)lua_tonumber(l, -1);
							lua_pop(l, 1);
					  }

					/* blue table is in the stack at index 7 */
					lua_pushnil(l);
					for( i=0; lua_next(l, 8) && i < RGB_RANGE; i++ )
						{
							ldata.b[i] = (unsigned char)lua_tonumber(l, -1);
							lua_pop(l, 1);
					  }

					/* at last, add the convolution kernel */
					res = tile_add_frame_data(tile, FRAME_LUT, w, h, data, &l);
			  }
			else if( !strcmp(mode, "xor") )
				res = tile_add_frame_data(tile, FRAME_XOR, w, h, data, NULL);
			else
			  return luaL_error(l, "Unknown frame type ");
		}


	/* and finished */
	if( res < 0 )
		{
			switch(res)
				{
					case ERR: return luaL_error(l, "Invalid tile id ");
					case ERR_BAD_FRAME_TYPE: return luaL_error(l, "Unknown frame type ");
					default: return luaL_error(l, "Unknown error ");
				}
		}
	else
		{
			lua_pushinteger(l, res);
			return 1;
		}

}




/* set the tile pixel mask */
int wrap_tile_set_pixel_mask(lua_State *l)
{
	/* tile pointer and frame index */
	tile *tile;
	int index;

	/* pixel data buffer */
	const unsigned char *data;
	size_t len;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	index = lua_tonumber(l, 2);
	data = lua_tolstring(l, 3, &len);

	tile_set_pixel_mask(tile, index, data);
	return 0;
}




/* set the tile pixel mask from a source frame */
int wrap_tile_set_pixel_mask_from(lua_State *l)
{
	/* tile pointer and frame index */
	tile *tile;
	int index;
	frame *frame;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	index = lua_tonumber(l, 2);

	frame = lua_touserdata(l, 3);
	if( !frame )
		return luaL_error(l, "Invalid frame ");

	tile_set_pixel_mask_from(tile, index, frame);
	return 0;
}




/* animate one tile */
int wrap_tile_animate(lua_State *l)
{
	tile *tile;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	tile_animate(tile);
	return 0;
}




/* reset one tile */
int wrap_tile_reset(lua_State *l)
{
	tile *tile;

	tile = lua_touserdata(l, 1);
	if( !tile )
		return luaL_error(l, "Invalid tile ");

	tile_reset(tile);
	return 0;
}


